@@ -2,7 +2,8 @@ use crate::annotation;
 use crate::core::*;
 use crate::encode::Encode;
 use crate::kw;
-use crate::parser::{Cursor, Parse, Parser, Result};
+use crate::lexer::{Lexer, Token, TokenKind};
+use crate::parser::{Parse, Parser, Result};
 use crate::token::*;
 use std::mem;
 
@@ -295,7 +296,7 @@ impl<'a> ExpressionParser<'a> {
                     }
                 }
                 Paren::None => {
-                    return Err(parser.error("expected to continue a folded instruction"))
+                    return Err(parser.error("expected to continue a folded instruction"));
                 }
             }
         }
@@ -441,7 +442,7 @@ macro_rules! instructions {
         }
 
         impl Encode for Instruction<'_> {
-            #[allow(non_snake_case)]
+            #[allow(non_snake_case, unused_lifetimes)]
             fn encode(&self, v: &mut Vec<u8>) {
                 match self {
                     $(
@@ -1187,6 +1188,20 @@ instructions! {
         I16x8RelaxedQ15mulrS: [0xfd, 0x111]: "i16x8.relaxed_q15mulr_s",
         I16x8RelaxedDotI8x16I7x16S: [0xfd, 0x112]: "i16x8.relaxed_dot_i8x16_i7x16_s",
         I32x4RelaxedDotI8x16I7x16AddS: [0xfd, 0x113]: "i32x4.relaxed_dot_i8x16_i7x16_add_s",
+
+        // Stack switching proposal
+        ContNew(Index<'a>)             : [0xe0] : "cont.new",
+        ContBind(ContBind<'a>)         : [0xe1] : "cont.bind",
+        Suspend(Index<'a>)             : [0xe2] : "suspend",
+        Resume(Resume<'a>)             : [0xe3] : "resume",
+        ResumeThrow(ResumeThrow<'a>)   : [0xe4] : "resume_throw",
+        Switch(Switch<'a>)             : [0xe5] : "switch",
+
+        // Wide arithmetic proposal
+        I64Add128   : [0xfc, 19] : "i64.add128",
+        I64Sub128   : [0xfc, 20] : "i64.sub128",
+        I64MulWideS : [0xfc, 21] : "i64.mul_wide_s",
+        I64MulWideU : [0xfc, 22] : "i64.mul_wide_u",
     }
 }
 
@@ -1194,11 +1209,12 @@ instructions! {
 // since big `*.wat` files will have a lot of these. This is a small ratchet to
 // make sure that this enum doesn't become larger than it already is, although
 // ideally it also wouldn't be as large as it is now.
-const _: () = {
+#[test]
+fn assert_instruction_not_too_large() {
     let size = std::mem::size_of::<Instruction<'_>>();
     let pointer = std::mem::size_of::<u64>();
-    assert!(size <= pointer * 10);
-};
+    assert!(size <= pointer * 11);
+}
 
 impl<'a> Instruction<'a> {
     pub(crate) fn needs_data_count(&self) -> bool {
@@ -1236,6 +1252,113 @@ impl<'a> Parse<'a> for BlockType<'a> {
     }
 }
 
+/// Extra information associated with the cont.bind instruction
+#[derive(Debug, Clone)]
+#[allow(missing_docs)]
+pub struct ContBind<'a> {
+    pub argument_index: Index<'a>,
+    pub result_index: Index<'a>,
+}
+
+impl<'a> Parse<'a> for ContBind<'a> {
+    fn parse(parser: Parser<'a>) -> Result<Self> {
+        Ok(ContBind {
+            argument_index: parser.parse()?,
+            result_index: parser.parse()?,
+        })
+    }
+}
+
+/// Extra information associated with the resume instruction
+#[derive(Debug, Clone)]
+#[allow(missing_docs)]
+pub struct Resume<'a> {
+    pub type_index: Index<'a>,
+    pub table: ResumeTable<'a>,
+}
+
+impl<'a> Parse<'a> for Resume<'a> {
+    fn parse(parser: Parser<'a>) -> Result<Self> {
+        Ok(Resume {
+            type_index: parser.parse()?,
+            table: parser.parse()?,
+        })
+    }
+}
+
+/// Extra information associated with the resume_throw instruction
+#[derive(Debug, Clone)]
+#[allow(missing_docs)]
+pub struct ResumeThrow<'a> {
+    pub type_index: Index<'a>,
+    pub tag_index: Index<'a>,
+    pub table: ResumeTable<'a>,
+}
+
+impl<'a> Parse<'a> for ResumeThrow<'a> {
+    fn parse(parser: Parser<'a>) -> Result<Self> {
+        Ok(ResumeThrow {
+            type_index: parser.parse()?,
+            tag_index: parser.parse()?,
+            table: parser.parse()?,
+        })
+    }
+}
+
+/// Extra information associated with the switch instruction
+#[derive(Debug, Clone)]
+#[allow(missing_docs)]
+pub struct Switch<'a> {
+    pub type_index: Index<'a>,
+    pub tag_index: Index<'a>,
+}
+
+impl<'a> Parse<'a> for Switch<'a> {
+    fn parse(parser: Parser<'a>) -> Result<Self> {
+        Ok(Switch {
+            type_index: parser.parse()?,
+            tag_index: parser.parse()?,
+        })
+    }
+}
+
+/// A representation of resume tables
+#[derive(Debug, Clone)]
+#[allow(missing_docs)]
+pub struct ResumeTable<'a> {
+    pub handlers: Vec<Handle<'a>>,
+}
+
+/// A representation of resume table entries
+#[derive(Debug, Clone)]
+#[allow(missing_docs)]
+pub enum Handle<'a> {
+    OnLabel { tag: Index<'a>, label: Index<'a> },
+    OnSwitch { tag: Index<'a> },
+}
+
+impl<'a> Parse<'a> for ResumeTable<'a> {
+    fn parse(parser: Parser<'a>) -> Result<Self> {
+        let mut handlers = Vec::new();
+        while parser.peek::<LParen>()? && parser.peek2::<kw::on>()? {
+            handlers.push(parser.parens(|p| {
+                p.parse::<kw::on>()?;
+                let tag: Index<'a> = p.parse()?;
+                if p.peek::<kw::switch>()? {
+                    p.parse::<kw::switch>()?;
+                    Ok(Handle::OnSwitch { tag })
+                } else {
+                    Ok(Handle::OnLabel {
+                        tag,
+                        label: p.parse()?,
+                    })
+                }
+            })?);
+        }
+        Ok(ResumeTable { handlers })
+    }
+}
+
 #[derive(Debug, Clone)]
 #[allow(missing_docs)]
 pub struct TryTable<'a> {
@@ -1364,7 +1487,7 @@ pub struct MemArg<'a> {
     ///
     /// This is not stored as a log, this is the actual alignment (e.g. 1, 2, 4,
     /// 8, etc).
-    pub align: u32,
+    pub align: u64,
     /// The offset, in bytes of this access.
     pub offset: u64,
     /// The memory index we're accessing
@@ -1372,12 +1495,8 @@ pub struct MemArg<'a> {
 }
 
 impl<'a> MemArg<'a> {
-    fn parse(parser: Parser<'a>, default_align: u32) -> Result<Self> {
-        fn parse_field<T>(
-            name: &str,
-            parser: Parser<'_>,
-            f: impl FnOnce(Cursor<'_>, &str, u32) -> Result<T>,
-        ) -> Result<Option<T>> {
+    fn parse(parser: Parser<'a>, default_align: u64) -> Result<Self> {
+        fn parse_field(name: &str, parser: Parser<'_>) -> Result<Option<u64>> {
             parser.step(|c| {
                 let (kw, rest) = match c.keyword()? {
                     Some(p) => p,
@@ -1391,35 +1510,34 @@ impl<'a> MemArg<'a> {
                     return Ok((None, c));
                 }
                 let num = &kw[1..];
-                let num = if let Some(stripped) = num.strip_prefix("0x") {
-                    f(c, stripped, 16)?
-                } else {
-                    f(c, num, 10)?
-                };
-
-                Ok((Some(num), rest))
-            })
-        }
-
-        fn parse_u32(name: &str, parser: Parser<'_>) -> Result<Option<u32>> {
-            parse_field(name, parser, |c, num, radix| {
-                u32::from_str_radix(num, radix).map_err(|_| c.error("i32 constant out of range"))
-            })
-        }
-
-        fn parse_u64(name: &str, parser: Parser<'_>) -> Result<Option<u64>> {
-            parse_field(name, parser, |c, num, radix| {
-                u64::from_str_radix(num, radix).map_err(|_| c.error("i64 constant out of range"))
+                let lexer = Lexer::new(num);
+                let mut pos = 0;
+                if let Ok(Some(
+                    token @ Token {
+                        kind: TokenKind::Integer(integer_kind),
+                        ..
+                    },
+                )) = lexer.parse(&mut pos)
+                {
+                    let int = token.integer(lexer.input(), integer_kind);
+                    let (s, base) = int.val();
+                    let value = u64::from_str_radix(s, base);
+                    return match value {
+                        Ok(n) => Ok((Some(n), rest)),
+                        Err(_) => Err(c.error("u64 constant out of range")),
+                    };
+                }
+                Err(c.error("expected u64 integer constant"))
             })
         }
 
         let memory = parser
             .parse::<Option<_>>()?
             .unwrap_or_else(|| Index::Num(0, parser.prev_span()));
-        let offset = parse_u64("offset", parser)?.unwrap_or(0);
-        let align = match parse_u32("align", parser)? {
+        let offset = parse_field("offset", parser)?.unwrap_or(0);
+        let align = match parse_field("align", parser)? {
             Some(n) if !n.is_power_of_two() => {
-                return Err(parser.error("alignment must be a power of two"))
+                return Err(parser.error("alignment must be a power of two"));
             }
             n => n.unwrap_or(default_align),
         };
@@ -1442,7 +1560,7 @@ pub struct LoadOrStoreLane<'a> {
 }
 
 impl<'a> LoadOrStoreLane<'a> {
-    fn parse(parser: Parser<'a>, default_align: u32) -> Result<Self> {
+    fn parse(parser: Parser<'a>, default_align: u64) -> Result<Self> {
         // This is sort of funky. The first integer we see could be the lane
         // index, but it could also be the memory index. To determine what it is
         // then if we see a second integer we need to look further.
